from .ops import NumpyOps, CupyOps


def get_ops(ops):
    if ops in ('numpy', 'cpu'):
        return NumpyOps()
    elif ops == ('cupy', 'gpu'):
        return CupyOps()
    else:
        return ops


def score_model(model, X_y):
    correct = 0
    total = 0
    X, y = zip(*X_y)
    scores = model.predict_batch(X)
    if isinstance(y, tuple) and (isinstance(y[0], tuple) or isinstance(y[0], list)):
        y = model.ops.asarray(model.ops.flatten(y), dtype='i')
    else:
        y = model.ops.asarray(y, dtype='i')
    for i, gold in enumerate(y):
        correct += scores[i].argmax() == gold
        total += 1
    return float(correct) / total


def partition(examples, split_size):
    examples = list(examples)
    numpy.random.shuffle(examples)
    n_docs = len(examples)
    split = int(n_docs * split_size)
    return examples[:split], examples[split:]


def minibatch(stream, batch_size=1000):
    batch = []
    for X in stream:
        batch.append(X)
        if len(batch) >= batch_size:
            yield batch
            batch = []
    if len(batch) != 0:
        yield batch



from ...ops import NumpyOps

import pytest
import numpy


@pytest.fixture
def ops():
    return NumpyOps()


def test_get_dropout_empty(ops):
    shape = (2,2)
    drop = 0.0
    mask = ops.get_dropout_mask(shape, drop)
    if drop <= 0.0:
        assert mask is None
    else:
        assert mask is not None


def test_get_dropout_not_empty(ops):
    shape = (2,2)
    drop = 0.1
    mask = ops.get_dropout_mask(shape, drop)
    if drop <= 0.0:
        assert mask is None
    else:
        assert mask is not None
    assert mask.shape == shape
    assert all(value >= 0 for value in mask.flatten())



import numpy
import pytest
from numpy.testing import assert_allclose

from ...ops import Ops
from ...vec2vec import Affine
from ...exceptions import ShapeError


class MockOps(Ops):
    def __init__(self):
        pass

    def allocate(self, shape, name=None):
        return numpy.zeros(shape)

    def allocate_pool(self, nr_weight, name=None):
        return DataPool(nr_weight)

    def affine(self, W_oi, b_o, input_bi):
        return numpy.ones((input_bi.shape[0], W_oi.shape[0]))

    def get_dropout(self, shape, drop):
        if drop <= 0:
            return None
        else:
            return numpy.ones(shape) * (1-drop)
    
    def batch_dot(self, x, y):
        return numpy.tensordot(x, y, axes=[[1], [1]])

    def batch_outer(self, x, y):
        return numpy.tensordot(x, y, axes=[[0], [0]])

    def xavier_uniform_init(self, W, inplace=True):
        pass


class DataPool(object):
    def __init__(self, nr_weight):
        self.data = numpy.zeros((nr_weight,))
        self.i = 0

    def allocate(self, nr_weight):
        assert self.i + nr_weight <= self.data.size
        data = self.data[self.i : self.i + nr_weight]
        self.i += nr_weight
        return data

    def allocate_shape(self, shape):
        return self.allocate(numpy.prod(shape)).reshape(shape)


@pytest.fixture
def ops():
    return MockOps()


@pytest.fixture
def model(ops):
    model = Affine(ops=ops, nr_out=10, nr_in=6)
    return model


def test_init(ops):
    model = Affine(ops=ops, nr_out=10, nr_in=6)
    assert model.nr_out == 10
    assert model.nr_in == 6
    assert model.W is not None
    assert model.b is not None
    assert isinstance(model.ops, MockOps)


def test_nr_weight(model):
    assert model.nr_weight == (model.nr_out * model.nr_in) + model.nr_out


def test_initialize_weights_no_pool(model):
    input_ = model.ops.allocate((5, 6))
    assert model.W.shape == (model.nr_out, model.nr_in)
    assert model.b.shape == (model.nr_out,)


def test_initialize_weights_pool(model):
    pool = DataPool(model.nr_weight)
    input_ = model.ops.allocate((5, 6))
    assert model.W.shape == (model.nr_out, model.nr_in)
    assert model.b.shape == (model.nr_out,)


def test_predict_batch_bias(model):
    input_ = model.ops.allocate((5, 6))
    model.b.fill(1)
    output = model.predict_batch(input_)
    assert output.shape == (5, 10)


def test_predict_batch_weights(model):
    input_ = model.ops.allocate((5, 6))
    output = model.predict_batch(input_)
    assert output.shape == (5, 10)
    assert all([val == 1. for val in output.flatten()]), output


def test_begin_update(model):
    input_ = model.ops.allocate((5, 6))
    output, finish_update = model.begin_update(input_)
    assert output.shape == (5, 10)
    assert all([val == 1. for val in output.flatten()]), output


def test_finish_update(model):
    seen_keys = set()
    def sgd(data, gradient, key=None, **kwargs):
        seen_keys.add(key)
        assert data.shape == gradient.shape

    model.name = 'model_name'
    input_ = model.ops.allocate((5, 6))
    output, finish_update = model.begin_update(input_)
    gradient = model.ops.allocate(output.shape)
    d_input = finish_update(gradient, sgd)
    assert d_input.shape == input_.shape
    assert ('W', 'model_name') in seen_keys
    assert ('b', 'model_name') in seen_keys


def test_predict_batch_not_batch(model):
    input_ = model.ops.allocate((6,))
    with pytest.raises(ShapeError):
        model.begin_update(input_)


def test_predict_update_dim_mismatch(model):
    input_ = model.ops.allocate((10, 5))
    with pytest.raises(ShapeError):
        model.begin_update(input_)


def test_begin_update_not_batch(model):
    input_ = model.ops.allocate((6,))
    with pytest.raises(ShapeError):
        model.begin_update(input_)


def test_begin_update_dim_mismatch(model):
    input_ = model.ops.allocate((10, 5))
    with pytest.raises(ShapeError):
        model.begin_update(input_)



import numpy
from cytoolz import concat


try:
    import cupy
except ImportError:
    cupy = None

try:
    import cytoolz as toolz
except ImportError:
    import toolz


class DataPool(object):
    def __init__(self, data):
        self.data = data
        self.i = 0

    def allocate(self, nr_weight):
        data = self.data[self.i : self.i + nr_weight]
        self.i += nr_weight
        return data

    def allocate_shape(self, shape):
        return self.allocate(numpy.prod(shape)).reshape(shape)


class Ops(object):
    xp = None

    def __init__(self, xp=None):
        if xp is not None:
            self.xp = xp

    def dropout(self, x, dropout, inplace=False):
        if dropout <= 0.0:
            return x, lambda func: func
        mask = self.get_dropout_mask(x.shape, dropout)
        def wrap_backprop(backprop):
            def finish_update(gradient, *args, **kwargs):
                return backprop(gradient * mask, *args, **kwargs)
            return finish_update
        if inplace:
            x *= mask
            return x, wrap_backprop
        else:
            return x * mask, wrap_backprop

    def flatten(self, X):
        return self.asarray(list(concat(X)))
 
    def unflatten(self, X, lengths):
        unflat = []
        for length in lengths:
            unflat.append(X[:length])
            X = X[length:]
        assert len(X) == 0
        assert len(unflat) == len(lengths)
        return unflat

    def get_dropout_mask(self, shape, drop):
        if drop <= 0.0:
            return None
        coinflips = self.xp.random.uniform(0., 1., shape)
        return (coinflips >= drop) / drop

    def allocate(self, shape, name=None):
        if isinstance(shape, int):
            shape = (shape,)
        nr_weight = numpy.prod(shape)
        return self.xp.zeros(shape, dtype='f')

    def allocate_pool(self, nr_weight, name=None):
        return DataPool(self.xp.zeros((nr_weight,), dtype='f'))

    def asarray(self, data, dtype='f'):
        return self.xp.asarray(data, dtype=dtype)

    def batch_dot(self, x, y):
        return self.xp.tensordot(x, y, axes=[[1], [1]])
   
    def batch_outer(self, x, y):
        return self.xp.tensordot(x, y, axes=[[0], [0]])

    def norm(self, x):
        return self.xp.sqrt((x * x).sum())

    def dot(self, x, y):
        return self.xp.dot(x, y)
    
    def affine(self, weights, bias, signal):
        return self.batch_dot(signal, weights) + bias

    def argmax(self, x, axis=-1):
        return self.xp.argmax(x, axis=axis)

    def softmax(self, x, inplace=False, axis=1):
        if x.ndim >= 3:
            raise NotImplementedError(
                "Softmax currently only supports 2d. ndim=%d" % x.ndim)
        shape = x.shape
        new_x = self.xp.zeros(shape=shape, dtype='f')
        for i in range(shape[0]):
            new_x[i] = self.xp.exp(x[i] - self.xp.max(x[i]))
            new_x[i] /= new_x[i].sum()
        if inplace:
            x[:] = new_x
            return x
        else:
            return new_x

    def expand_dims(self, a, axis=-1):
        return self.xp.expand_dims(a, axis=axis)

    def clip_low(self, x, value, inplace=False):
        if inplace:
            return self.xp.maximum(x, value, out=x)
        else:
            return self.xp.maximum(x, value)

    def take_which(self, x, which, axis=-1):
        output = self.allocate(which.shape)
        for i in range(x.shape[axis]):
            output += x[:,:,i] * (which == i)
        return output

    def xavier_uniform_init(self, W, inplace=True):
        scale = self.xp.sqrt(2. / (W.shape[0] + W.shape[1]))
        if inplace:
            W[:] = self.xp.random.uniform(-scale, scale, W.shape)
            return W
        else:
            return self.xp.random.uniform(-scale, scale, W.shape)


class NumpyOps(Ops):
    xp = numpy


class CupyOps(Ops):
    xp = cupy



from .base import Model
from .exceptions import ShapeError


class Affine(Model):
    name = 'affine'
    nr_out = None
    nr_in = None
    data = None

    @property
    def is_initialized(self):
        return self.W is not None

    @property
    def input_shape(self):
        return (self.nr_in,)

    @property
    def output_shape(self):
        return (self.nr_out,)

    @property
    def nr_weight(self):
        return (self.nr_out * self.nr_in) + self.nr_out

    def setup(self, *args, **kwargs):
        self.W = None
        if 'W' in kwargs:
            self.nr_out = kwargs.get('W').shape[0]
            self.nr_in = kwargs.get('W').shape[1]
        if self.nr_out is not None and self.nr_in is not None:
            self.set_weights(initialize=True)
            self.set_gradient()
        if 'W' in kwargs:
            self.W[:] = kwargs.get('W')
        if 'b' in kwargs:
            self.b[:] = kwargs.get('b')

    def set_weights(self, data=None, initialize=True, example=None):
        if example is not None:
            self.nr_in = example.shape[-1]
        if data is None:
            if self.data is None:
                self.data = self.ops.allocate_pool(self.nr_weight,
                                name=(self.name, 'pool'))
            data = self.data
        self.W = data.allocate_shape((self.nr_out, self.nr_in))
        self.b = data.allocate_shape((self.nr_out,))
        if initialize:
            self.ops.xavier_uniform_init(self.W, inplace=True)

    def set_gradient(self, data=None, initialize=False):
        if data is None:
            self.d_data = self.ops.allocate_pool(self.nr_weight,
                            name=(self.name, 'pool'))
        else:
            self.d_data = data
        self.d_W = self.d_data.allocate_shape((self.nr_out, self.nr_in))
        self.d_b = self.d_data.allocate_shape((self.nr_out,))

    def predict_batch(self, input_BI):
        return self.ops.affine(self.W, self.b, input_BI)

    def begin_update(self, input_BI, dropout=0.0):
        self.check_shape(input_BI, True)
        output_BO = self.predict_batch(input_BI)
        output_BO, bp_dropout = self.ops.dropout(output_BO, dropout)
        return output_BO, bp_dropout(self._get_finish_update(input_BI))
   
    def _get_finish_update(self, acts_BI):
        def finish_update(d_acts_BO, optimizer=None, **kwargs):
            self.d_b += d_acts_BO.sum(axis=0)
            self.d_W += self.ops.batch_outer(d_acts_BO, acts_BI)
            if optimizer is not None and self.data is not None:
                optimizer(self.W, self.d_W, key=('W', self.name), **kwargs)
                optimizer(self.b, self.d_b, key=('b', self.name), **kwargs)
            return self.ops.batch_dot(d_acts_BO, self.W.T)
        return finish_update


class ReLu(Affine):
    name = 'relu'
    def predict_batch(self, input_BI):
        output_BO = Affine.predict_batch(self, input_BI)
        return self.ops.xp.maximum(output_BO, 0., out=output_BO)

    def begin_update(self, input_BI, dropout=0.0):
        output_BO, finish_affine = Affine.begin_update(self, input_BI)
        def finish_update(gradient, *args, **kwargs):
            return finish_affine(gradient * (output_BO > 0), *args, **kwargs)
        return output_BO, finish_update


class Softmax(Affine):
    name = 'softmax'

    def set_weights(self, example=None, data=None, initialize=True):
        Affine.set_weights(self, example=example, data=data, initialize=False)

    def predict_batch(self, input_bi):
        output_bo = Affine.predict_batch(self, input_bi)
        return self.ops.softmax(output_bo, axis=-1)

    def begin_update(self, input_BI, dropout=0.0):
        return Affine.begin_update(self, input_BI, dropout=0.0)


class Maxout(Affine):
    name = 'maxout'
    nr_piece = 3

    @property
    def nr_weight(self):
        return self.nr_piece * ((self.nr_out * self.nr_in) + self.nr_out)

    def set_weights(self, data=None, initialize=True, example=None):
        if example is not None:
            self.nr_in = example.shape[-1]
        if data is None:
            if self.data is None:
                self.data = self.ops.allocate_pool(self.nr_weight,
                                name=(self.name, 'pool'))
            data = self.data
        self.W = data.allocate_shape((self.nr_out, self.nr_piece, self.nr_in))
        self.b = data.allocate_shape((self.nr_out, self.nr_piece))
        if initialize:
            for i in range(self.nr_piece):
                self.ops.xavier_uniform_init(self.W[:, i], inplace=True)

    def set_gradient(self, data=None, initialize=False):
        if data is None:
            self.d_data = self.ops.allocate_pool(self.nr_weight,
                            name=(self.name, 'pool'))
        else:
            self.d_data = data
        self.d_W = self.d_data.allocate_shape((self.nr_out, self.nr_piece, self.nr_in))
        self.d_b = self.d_data.allocate_shape((self.nr_out, self.nr_piece))


    def predict_batch(self, input_bi):
        acts_bop = self.ops.xp.tensordot(input_bi, self.W, axes=[[1], [-1]])
        acts_bop += self.b
        which_bo = self.ops.argmax(acts_bop, axis=-1)
        return _take_which(self.ops, acts_bop, which_bo)

    def begin_update(self, input_BI, dropout=0.0):
        W_OCI = self.W
        b_OC = self.b
        output_BOC = self.ops.xp.tensordot(input_BI, W_OCI, axes=[[1], [-1]])
        output_BOC += b_OC
        which_BO = self.ops.argmax(output_BOC, axis=-1)
        best_BO = _take_which(self.ops, output_BOC, which_BO)
        best_BO, bp_dropout = self.ops.dropout(best_BO, dropout, inplace=True)
        finish_update = self._get_finish_update(input_BI, which_BO)
        return best_BO, bp_dropout(finish_update)

    def _get_finish_update(self, acts_BI, which_BO):
        def finish_update(d_acts_BO, optimizer=None, **kwargs):
            d_acts_BOP = self.ops.allocate((d_acts_BO.shape[0], self.nr_out,
                                           self.nr_piece))
            for i in range(self.nr_piece):
                d_acts_BOP[:, :, i] += d_acts_BO * (which_BO == i)
            self.d_b += d_acts_BOP.sum(axis=0)
            self.d_W += self.ops.xp.tensordot(d_acts_BOP, acts_BI, axes=[[0], [0]])
            # Bop,opi->Bi
            d_acts_BI = self.ops.xp.tensordot(d_acts_BOP, self.W, axes=[[1,2], [0, 1]])
            return d_acts_BI
        return finish_update


def _take_which(ops, x, which, axis=-1):
    output = ops.allocate(which.shape)
    for i in range(x.shape[axis]):
        output += x[:, :, i] * (which == i)
    return output
 
